package immibis.microblocks;

import immibis.core.api.util.Dir;
import immibis.microblocks.api.EnumAxisPosition;
import immibis.microblocks.api.EnumPartClass;
import immibis.microblocks.api.EnumPosition;
import immibis.microblocks.api.Part;
import immibis.microblocks.api.PartType;
import net.minecraft.block.Block;
import net.minecraft.client.renderer.RenderBlocks;
import net.minecraft.client.renderer.RenderGlobal;
import net.minecraft.client.renderer.Tessellator;
import net.minecraft.util.AxisAlignedBB;
import net.minecraftforge.client.ForgeHooksClient;

import org.lwjgl.opengl.GL11;

public class MicroblockRenderer {
	private static double u, v;
	private static int[] textures;
	private static double size;
	private static void setTex(int tex)
	{
		u = (tex % 16) / 16.0;
		v = ((tex / 16) + 1) / 16.0;
	}
	
	private static void vertXY(Tessellator t, double x, double y, double z) {	
		t.addVertexWithUV(x, y, z, u+x/16.0, v-y/16.0);
	}
	
	private static void vertXZ(Tessellator t, double x, double y, double z) {	
		t.addVertexWithUV(x, y, z, u+x/16.0, v-z/16.0);
	}
	
	private static void vertYZ(Tessellator t, double x, double y, double z) {	
		t.addVertexWithUV(x, y, z, u+z/16.0, v-y/16.0);
	}
	
	public static void renderPartPreview(RenderGlobal render, EnumPosition pos, PartType p) {
		size = p.size; textures = p.textures;
		GL11.glEnable(GL11.GL_BLEND);
		GL11.glColor4f(1.0f, 1.0f, 1.0f, 0.5f);
		
		boolean customTex = !p.texfile.equals("/terrain.png");
		
		if(customTex)
			ForgeHooksClient.bindTexture(p.texfile, 0);

		Tessellator t = Tessellator.instance;
		t.startDrawingQuads();
		
		renderQuads(t, pos, p);
		
		t.draw();
		
		if(customTex)
			ForgeHooksClient.unbindTexture();
		
		GL11.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		GL11.glDisable(GL11.GL_BLEND);
	}
	
	private static final double HOLLOW_EDGE_SIZE = 0.25;
	
	private static void renderHollowPanelYZ(Tessellator t, AxisAlignedBB aabb) {
		setTex(textures[Dir.NX]);
		t.setNormal(-1, 0, 0);
		
		vertYZ(t, aabb.minX, 0+HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, 1, 0);
		vertYZ(t, aabb.minX, 0, 0);
		
		vertYZ(t, aabb.minX, 0, 1);
		vertYZ(t, aabb.minX, 1, 1);
		vertYZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		
		vertYZ(t, aabb.minX, 0, 1);
		vertYZ(t, aabb.minX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, 0, 0);
		
		vertYZ(t, aabb.minX, 1, 0);
		vertYZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.minX, 1, 1);
		
		setTex(textures[Dir.PX]);
		t.setNormal(1, 0, 0);
		
		vertYZ(t, aabb.maxX, 0, 0);
		vertYZ(t, aabb.maxX, 1, 0);
		vertYZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		
		vertYZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, 1, 1);
		vertYZ(t, aabb.maxX, 0, 1);
		
		vertYZ(t, aabb.maxX, 0, 0);
		vertYZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, 0+HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, 0, 1);
		
		vertYZ(t, aabb.maxX, 1, 1);
		vertYZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertYZ(t, aabb.maxX, 1, 0);
		
		setTex(textures[Dir.NY]);
		t.setNormal(0, -1, 0);
		vertXZ(t, aabb.minX, 0, 0);
		vertXZ(t, aabb.maxX, 0, 0);
		vertXZ(t, aabb.maxX, 0, 1);
		vertXZ(t, aabb.minX, 0, 1);
		
		vertXZ(t, aabb.minX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.maxX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.minX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.PY]);
		t.setNormal(0, 1, 0);
		vertXZ(t, aabb.minX, 1, 1);
		vertXZ(t, aabb.maxX, 1, 1);
		vertXZ(t, aabb.maxX, 1, 0);
		vertXZ(t, aabb.minX, 1, 0);
		
		vertXZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.NZ]);
		t.setNormal(0, 0, -1);
		vertXY(t, aabb.minX, 0, 0);
		vertXY(t, aabb.minX, 1, 0);
		vertXY(t, aabb.maxX, 1, 0);
		vertXY(t, aabb.maxX, 0, 0);
		
		vertXY(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.minX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.maxX, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.PZ]);
		t.setNormal(0, 0, 1);
		vertXY(t, aabb.minX, 1, 1);
		vertXY(t, aabb.minX, 0, 1);
		vertXY(t, aabb.maxX, 0, 1);
		vertXY(t, aabb.maxX, 1, 1);
		
		vertXY(t, aabb.minX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.minX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.maxX, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, aabb.maxX, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE);
	}
	
	private static void renderHollowPanelXZ(Tessellator t, AxisAlignedBB aabb) {
		setTex(textures[Dir.NY]);
		t.setNormal(0, -1, 0);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, 0, aabb.minY, 1);
		vertXZ(t, 0, aabb.minY, 0);
		
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertXZ(t, 1, aabb.minY, 0);
		vertXZ(t, 1, aabb.minY, 1);
		
		vertXZ(t, 0, aabb.minY, 0);
		vertXZ(t, 1, aabb.minY, 0);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, 1, aabb.minY, 1);
		vertXZ(t, 0, aabb.minY, 1);
		
		setTex(textures[Dir.PY]);
		t.setNormal(0, 1, 0);
		vertXZ(t, 0, aabb.maxY, 0);
		vertXZ(t, 0, aabb.maxY, 1);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		
		vertXZ(t, 1, aabb.maxY, 1);
		vertXZ(t, 1, aabb.maxY, 0);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertXZ(t, 1, aabb.maxY, 0);
		vertXZ(t, 0, aabb.maxY, 0);
		
		vertXZ(t, 0, aabb.maxY, 1);
		vertXZ(t, 1, aabb.maxY, 1);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		vertXZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.NX]);
		t.setNormal(-1, 0, 0);
		vertYZ(t, 0, aabb.minY, 0);
		vertYZ(t, 0, aabb.minY, 1);
		vertYZ(t, 0, aabb.maxY, 1);
		vertYZ(t, 0, aabb.maxY, 0);
		
		vertYZ(t, HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertYZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertYZ(t, HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.PX]);
		t.setNormal(1, 0, 0);
		vertYZ(t, 1, aabb.minY, 1);
		vertYZ(t, 1, aabb.minY, 0);
		vertYZ(t, 1, aabb.maxY, 0);
		vertYZ(t, 1, aabb.maxY, 1);
		
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.NZ]);
		t.setNormal(0, 0, -1);
		vertXY(t, 0, aabb.minY, 0);
		vertXY(t, 0, aabb.maxY, 0);
		vertXY(t, 1, aabb.maxY, 0);
		vertXY(t, 1, aabb.minY, 0);
		
		vertXY(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertXY(t, HOLLOW_EDGE_SIZE, aabb.maxY, HOLLOW_EDGE_SIZE);
		vertXY(t, HOLLOW_EDGE_SIZE, aabb.minY, HOLLOW_EDGE_SIZE);
		
		setTex(textures[Dir.PZ]);
		t.setNormal(0, 0, 1);
		vertXY(t, 1, aabb.minY, 1);
		vertXY(t, 1, aabb.maxY, 1);
		vertXY(t, 0, aabb.maxY, 1);
		vertXY(t, 0, aabb.minY, 1);
		
		vertXY(t, HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, aabb.maxY, 1-HOLLOW_EDGE_SIZE);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, aabb.minY, 1-HOLLOW_EDGE_SIZE);
	}
		
	private static void renderHollowPanelXY(Tessellator t, AxisAlignedBB aabb) {
		setTex(textures[Dir.NZ]);
		t.setNormal(0, 0, -1);
		vertXY(t, 0, 0, aabb.minZ);
		vertXY(t, 0, 1, aabb.minZ);
		vertXY(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXY(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		
		vertXY(t, 1, 1, aabb.minZ);
		vertXY(t, 1, 0, aabb.minZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		
		vertXY(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXY(t, 1, 0, aabb.minZ);
		vertXY(t, 0, 0, aabb.minZ);
		
		vertXY(t, 0, 1, aabb.minZ);
		vertXY(t, 1, 1, aabb.minZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXY(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		
		setTex(textures[Dir.PZ]);
		t.setNormal(0, 0, 1);
		vertXY(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, 0, 1, aabb.maxZ);
		vertXY(t, 0, 0, aabb.maxZ);
		
		vertXY(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, 1, 0, aabb.maxZ);
		vertXY(t, 1, 1, aabb.maxZ);
		
		vertXY(t, 0, 0, aabb.maxZ);
		vertXY(t, 1, 0, aabb.maxZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		
		vertXY(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXY(t, 1, 1, aabb.maxZ);
		vertXY(t, 0, 1, aabb.maxZ);
		
		setTex(textures[Dir.NX]);
		t.setNormal(-1, 0, 0);
		vertYZ(t, 0, 1, aabb.minZ);
		vertYZ(t, 0, 0, aabb.minZ);
		vertYZ(t, 0, 0, aabb.maxZ);
		vertYZ(t, 0, 1, aabb.maxZ);
		
		vertYZ(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		vertYZ(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		vertYZ(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertYZ(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		
		setTex(textures[Dir.PX]);
		t.setNormal(1, 0, 0);
		vertYZ(t, 1, 0, aabb.minZ);
		vertYZ(t, 1, 1, aabb.minZ);
		vertYZ(t, 1, 1, aabb.maxZ);
		vertYZ(t, 1, 0, aabb.maxZ);
		
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		vertYZ(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		
		setTex(textures[Dir.NY]);
		t.setNormal(0, -1, 0);
		vertXZ(t, 1, 0, aabb.minZ);
		vertXZ(t, 1, 0, aabb.maxZ);
		vertXZ(t, 0, 0, aabb.maxZ);
		vertXZ(t, 0, 0, aabb.minZ);
		
		vertXZ(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXZ(t, HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, HOLLOW_EDGE_SIZE, aabb.minZ);
		
		setTex(textures[Dir.PY]);
		t.setNormal(0, 1, 0);
		vertXZ(t, 0, 1, aabb.minZ);
		vertXZ(t, 0, 1, aabb.maxZ);
		vertXZ(t, 1, 1, aabb.maxZ);
		vertXZ(t, 1, 1, aabb.minZ);
		
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
		vertXZ(t, 1-HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXZ(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.maxZ);
		vertXZ(t, HOLLOW_EDGE_SIZE, 1-HOLLOW_EDGE_SIZE, aabb.minZ);
	}
	
	protected static void renderQuads(Tessellator t, EnumPosition pos, PartType pt) {
		AxisAlignedBB aabb = Part.getBoundingBoxFromPool(pos, size);
		
		if(pt.clazz == EnumPartClass.HollowPanel)
		{
			if(pos.x != EnumAxisPosition.Span)
				renderHollowPanelYZ(t, aabb);
			else if(pos.y != EnumAxisPosition.Span)
				renderHollowPanelXZ(t, aabb);
			else if(pos.z != EnumAxisPosition.Span)
				renderHollowPanelXY(t, aabb);
			return;
		}
		
		renderAABB(t, aabb, textures, null, 0, 0, 0, null);
	}
	
	static double interp(double a, double b, double i) {
		return a + (b - a) * i;
	}
	static int rgb_component(double a) {
		if(a < 0)
			return 0;
		if(a > 1)
			return 255;
		return (int)(255 * a);
	}
	static int rgb(double r, double g, double b) {
		return (rgb_component(r) << 16) | (rgb_component(g) << 8) | rgb_component(b);
	}
	static int rgb_r(int a) {return (a >> 16) & 255;}
	static int rgb_g(int a) {return (a >> 8) & 255;}
	static int rgb_b(int a) {return a & 255;}
	static int interpRGB(int a, int b, double i) {
		return rgb(interp(rgb_r(a), rgb_r(b), i)/255.0, interp(rgb_g(a), rgb_g(b), i)/255.0, interp(rgb_b(a), rgb_b(b), i)/255.0);
	}
	static int scaleRGB(int a, double sc) {
		sc /= 255;
		return rgb(rgb_r(a)*sc, rgb_g(a)*sc, rgb_b(a)*sc);
	}
	static int interpBrightness(int a, int b, double i) {
		int AH = a >> 16;
		int AL = a & 65535;
		int BH = b >> 16;
		int BL = b & 65535;
		int H = (int)(AH + (BH - AH) * i);
		int L = (int)(AL + (BL - AL) * i);
		return (H << 16) | L;
	}
	
	// colour and brightness values for the corners of the face currently being rendered
	private static int colNN, colPN, colNP, colPP;
	private static int briNN, briPN, briNP, briPP;
	
	static void setColourAndBrightness(Tessellator t, double x, double y) {
		int col = interpRGB(interpRGB(colNN, colPN, x), interpRGB(colNP, colPP, x), y);
		int bri = interpBrightness(interpBrightness(briNN, briPN, x), interpBrightness(briNP, briPP, x), y);
		t.setColorOpaque_I(col);
		t.setBrightness(bri);
	}
	
	static void renderAABB(Tessellator t, AxisAlignedBB aabb, int[] textures, RenderBlocks rb, int bx, int by, int bz, Block par1Block) {
		int tex;
		
		// block-relative AABB
		AxisAlignedBB rbb = aabb.getOffsetBoundingBox(-bx, -by, -bz);
		
		// texture offsets from fractional coordinates
		double txmin = (((aabb.minX % 1) + 1) % 1) / 16.0;
		double txmax = (((aabb.maxX % 1) + 1) % 1) / 16.0;
		double tymin = (((aabb.minY % 1) + 1) % 1) / 16.0;
		double tymax = (((aabb.maxY % 1) + 1) % 1) / 16.0;
		double tzmin = (((aabb.minZ % 1) + 1) % 1) / 16.0;
		double tzmax = (((aabb.maxZ % 1) + 1) % 1) / 16.0;
		
		boolean usingAO = (rb != null) && false;
		
		if(!usingAO) {
			colNN = colPN = colNP = colPP = 0x00FFFFFF;
			briNN = briPN = briNP = briPP = 0x00F000F0;
		}
		
		if(txmax == 0) txmax = 1/16.0;
		if(tymax == 0) tymax = 1/16.0;
		if(tzmax == 0) tzmax = 1/16.0;
		
		float colourR = 1, colourG = 1, colourB = 1;
		
		int var19 = rb == null ? 0 : par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx, by, bz);
		
		if(rb != null)
			briNN = briNP = briPP = briPN = var19;
		
		tex = textures[Dir.NZ]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		t.setNormal(0, 0, -1);
		colNN = colPN = colNP = colPP = rgb(colourR*0.8, colourG*0.8, colourB*0.8);
		setColourAndBrightness(t, rbb.minX, rbb.minY);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.minZ, u+txmin, v-tymin);
		setColourAndBrightness(t, rbb.minX, rbb.maxY);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.minZ, u+txmin, v-tymax);
		setColourAndBrightness(t, rbb.maxX, rbb.maxY);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.minZ, u+txmax, v-tymax);
		setColourAndBrightness(t, rbb.maxX, rbb.minY);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.minZ, u+txmax, v-tymin);
		
		tex = textures[Dir.NY]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		if(usingAO)
		{
			if(rbb.minY <= 0) by--;
			
            rb.aoBrightnessXYNN = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx - 1, by, bz);
            rb.aoBrightnessYZNN = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx, by, bz - 1);
            rb.aoBrightnessYZNP = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx, by, bz + 1);
            rb.aoBrightnessXYPN = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx + 1, by, bz);
            rb.aoLightValueScratchXYNN = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx - 1, by, bz);
            rb.aoLightValueScratchYZNN = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx, by, bz - 1);
            rb.aoLightValueScratchYZNP = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx, by, bz + 1);
            rb.aoLightValueScratchXYPN = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx + 1, by, bz);

            if (!rb.aoGrassXYZCNN && !rb.aoGrassXYZNNC)
            {
                rb.aoLightValueScratchXYZNNN = rb.aoLightValueScratchXYNN;
                rb.aoBrightnessXYZNNN = rb.aoBrightnessXYNN;
            }
            else
            {
                rb.aoLightValueScratchXYZNNN = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx - 1, by, bz - 1);
                rb.aoBrightnessXYZNNN = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx - 1, by, bz - 1);
            }

            if (!rb.aoGrassXYZCNP && !rb.aoGrassXYZNNC)
            {
                rb.aoLightValueScratchXYZNNP = rb.aoLightValueScratchXYNN;
                rb.aoBrightnessXYZNNP = rb.aoBrightnessXYNN;
            }
            else
            {
                rb.aoLightValueScratchXYZNNP = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx - 1, by, bz + 1);
                rb.aoBrightnessXYZNNP = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx - 1, by, bz + 1);
            }

            if (!rb.aoGrassXYZCNN && !rb.aoGrassXYZPNC)
            {
                rb.aoLightValueScratchXYZPNN = rb.aoLightValueScratchXYPN;
                rb.aoBrightnessXYZPNN = rb.aoBrightnessXYPN;
            }
            else
            {
                rb.aoLightValueScratchXYZPNN = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx + 1, by, bz - 1);
                rb.aoBrightnessXYZPNN = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx + 1, by, bz - 1);
            }

            if (!rb.aoGrassXYZCNP && !rb.aoGrassXYZPNC)
            {
                rb.aoLightValueScratchXYZPNP = rb.aoLightValueScratchXYPN;
                rb.aoBrightnessXYZPNP = rb.aoBrightnessXYPN;
            }
            else
            {
                rb.aoLightValueScratchXYZPNP = par1Block.getAmbientOcclusionLightValue(rb.blockAccess, bx + 1, by, bz + 1);
                rb.aoBrightnessXYZPNP = par1Block.getMixedBrightnessForBlock(rb.blockAccess, bx + 1, by, bz + 1);
            }
            
            if(rbb.minY <= 0) by++;

            float var9 = (rb.aoLightValueScratchXYZNNP + rb.aoLightValueScratchXYNN + rb.aoLightValueScratchYZNP + rb.aoLightValueYNeg) / 4.0F;
            float var12 = (rb.aoLightValueScratchYZNP + rb.aoLightValueYNeg + rb.aoLightValueScratchXYZPNP + rb.aoLightValueScratchXYPN) / 4.0F;
            float var11 = (rb.aoLightValueYNeg + rb.aoLightValueScratchYZNN + rb.aoLightValueScratchXYPN + rb.aoLightValueScratchXYZPNN) / 4.0F;
            float var10 = (rb.aoLightValueScratchXYNN + rb.aoLightValueScratchXYZNNN + rb.aoLightValueYNeg + rb.aoLightValueScratchYZNN) / 4.0F;
            briNP = rb.getAoBrightness(rb.aoBrightnessXYZNNP, rb.aoBrightnessXYNN, rb.aoBrightnessYZNP, var19);
            briPP = rb.getAoBrightness(rb.aoBrightnessYZNP, rb.aoBrightnessXYZPNP, rb.aoBrightnessXYPN, var19);
            briPN = rb.getAoBrightness(rb.aoBrightnessYZNN, rb.aoBrightnessXYPN, rb.aoBrightnessXYZPNN, var19);
            briNN = rb.getAoBrightness(rb.aoBrightnessXYNN, rb.aoBrightnessXYZNNN, rb.aoBrightnessYZNN, var19);
            colNN = colPN = colNP = colPP = rgb(colourR*0.5, colourG*0.5, colourB*0.5);
            colNP = scaleRGB(colNP, var9);
            colNN = scaleRGB(colNN, var10);
            colPN = scaleRGB(colPN, var11);
            colPP = scaleRGB(colPP, var12);
		} else
			colNN = colPN = colNP = colPP = rgb(colourR*0.5, colourG*0.5, colourB*0.5);
		t.setNormal(0, -1, 0);
		setColourAndBrightness(t, rbb.maxX, rbb.minZ);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.minZ, u+txmax, v-tzmin);
		setColourAndBrightness(t, rbb.maxX, rbb.maxZ);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.maxZ, u+txmax, v-tzmax);
		setColourAndBrightness(t, rbb.minX, rbb.maxZ);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.maxZ, u+txmin, v-tzmax);
		setColourAndBrightness(t, rbb.minX, rbb.minZ);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.minZ, u+txmin, v-tzmin);
		
		tex = textures[Dir.NX]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		t.setNormal(-1, 0, 0);
		colNN = colPN = colNP = colPP = rgb(colourR*0.6, colourG*0.6, colourB*0.6);
		setColourAndBrightness(t, rbb.minY, rbb.minZ);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.minZ, u+tzmin, v-tymin);
		setColourAndBrightness(t, rbb.minY, rbb.maxZ);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.maxZ, u+tzmax, v-tymin);
		setColourAndBrightness(t, rbb.maxY, rbb.maxZ);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.maxZ, u+tzmax, v-tymax);
		setColourAndBrightness(t, rbb.maxY, rbb.minZ);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.minZ, u+tzmin, v-tymax);
		
		tex = textures[Dir.PZ]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		t.setNormal(0, 0, 1);
		colNN = colPN = colNP = colPP = rgb(colourR*0.8, colourG*0.8, colourB*0.8);
		setColourAndBrightness(t, rbb.maxX, rbb.minY);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.maxZ, u+txmax, v-tymin);
		setColourAndBrightness(t, rbb.maxX, rbb.maxY);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.maxZ, u+txmax, v-tymax);
		setColourAndBrightness(t, rbb.minX, rbb.maxY);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.maxZ, u+txmin, v-tymax);
		setColourAndBrightness(t, rbb.minX, rbb.minY);
		t.addVertexWithUV(aabb.minX, aabb.minY, aabb.maxZ, u+txmin, v-tymin);
		
		tex = textures[Dir.PY]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		t.setNormal(0, 1, 0);
		colNN = colPN = colNP = colPP = rgb(colourR*1.0, colourG*1.0, colourB*1.0);
		setColourAndBrightness(t, rbb.minX, rbb.minZ);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.minZ, u+txmin, v-tzmin);
		setColourAndBrightness(t, rbb.minX, rbb.maxZ);
		t.addVertexWithUV(aabb.minX, aabb.maxY, aabb.maxZ, u+txmin, v-tzmax);
		setColourAndBrightness(t, rbb.maxX, rbb.maxZ);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.maxZ, u+txmax, v-tzmax);
		setColourAndBrightness(t, rbb.maxX, rbb.minZ);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.minZ, u+txmax, v-tzmin);
		
		tex = textures[Dir.PX]; u = (tex % 16) / 16.0; v = ((tex / 16) + 1) / 16.0;
		t.setNormal(1, 0, 0);
		colNN = colPN = colNP = colPP = rgb(colourR*0.6, colourG*0.6, colourB*0.6);
		setColourAndBrightness(t, rbb.maxY, rbb.minZ);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.minZ, u+tzmin, v-tymax);
		setColourAndBrightness(t, rbb.maxY, rbb.maxZ);
		t.addVertexWithUV(aabb.maxX, aabb.maxY, aabb.maxZ, u+tzmax, v-tymax);
		setColourAndBrightness(t, rbb.minY, rbb.maxZ);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.maxZ, u+tzmax, v-tymin);
		setColourAndBrightness(t, rbb.minY, rbb.minZ);
		t.addVertexWithUV(aabb.maxX, aabb.minY, aabb.minZ, u+tzmin, v-tymin);
	}
	
	// renders centred on 0.5,0.5,0.5
	public static void renderPartInv(RenderBlocks render, PartType p) {
		size = p.size; textures = p.textures;
		GL11.glDisable(GL11.GL_BLEND);
		GL11.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		GL11.glPushMatrix();
		
		EnumPosition pos;
		switch(p.clazz)
		{
		case Centre:
			pos = EnumPosition.Centre;
			break;
		case Panel: case HollowPanel:
			pos = EnumPosition.FaceNZ;
			GL11.glTranslatef(0, 0, 0.5f-(float)size/2);
			break;
		case Strip:
			pos = EnumPosition.EdgeNXNZ;
			GL11.glTranslatef(0.5f-(float)size/2, 0, 0.5f-(float)size/2);
			break;
		case Corner:
			pos = EnumPosition.CornerNXNYNZ;
			GL11.glTranslatef(0.5f-(float)size/2, 0.5f-(float)size/2, 0.5f-(float)size/2);
			break;
		default:
			pos = EnumPosition.Centre;
		}
		
		Tessellator t = Tessellator.instance;
		t.startDrawingQuads();
		
		renderQuads(t, pos, p);
		
		t.draw();
		GL11.glPopMatrix();
		GL11.glColor4f(1.0f, 1.0f, 1.0f, 1.0f);
		GL11.glDisable(GL11.GL_BLEND);
	}
}
