package immibis.microblocks.coremod;

import immibis.microblocks.coremod.OptionsFile.BooleanOption;
import immibis.microblocks.coremod.OptionsFile.ItemAndMetaListOption;
import immibis.microblocks.coremod.OptionsFile.StringListOption;
import immibis.microblocks.coremod.OptionsFile.ItemListOption.ItemID;

import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Properties;

import com.google.common.collect.ImmutableList;

public class CoreModOptions {
	public static boolean autoDetectCuttableBlocks;
	public static List<ItemID> manualCuttableBlocks;
	public static boolean enableNEI;
	
	public static final ImmutableList<String> DEFAULT_BLOCK_CLASSES = ImmutableList.of(
		"buildcraft.transport.BlockGenericPipe",
		"thermalexpansion.transport.block.BlockConduit",
		"ic2.core.block.wiring.BlockCable",
		"appeng.common.AppEngMultiBlock"
	);
	public static final ImmutableList<String> DEFAULT_TILE_CLASSES = ImmutableList.of(
		"buildcraft.transport.TileGenericPipe",
		"thermalexpansion.transport.tileentity.TileConduitEnergy",
		"thermalexpansion.transport.tileentity.TileConduitLiquid",
		"ic2.core.block.wiring.TileEntityCable",
		"appeng.me.tile.TileCable",
		"appeng.me.tile.TileDarkCable",
		"appeng.me.tile.TileInputCable",
		"appeng.me.tile.TileOutputCable",
		"appeng.me.tile.TileLevelEmitter",
		"appeng.me.tile.TileStorageBus"
	);
	
	private static boolean loaded;
	private static File configFile;
	private static File configDir;
	
	private static OptionsFile of;
	
	private static void updateConfig() throws IOException {
		File oldConfigFile = new File(configDir, "immibis-coremod.cfg");
		if(oldConfigFile.exists()) {
			Properties props = new Properties();
			
			try {
				FileReader fr = new FileReader(oldConfigFile);
				try {
					props.load(fr);
				} finally {
					fr.close();
				}
			} catch(IOException ex) {
				throw new RuntimeException(ex);
			}
			
			StringListOption blockClasses = new StringListOption();
			StringListOption tileClasses = new StringListOption();
			BooleanOption autoDetect = new BooleanOption();
			
			OptionsFile of = new OptionsFile();
			of.addOption("blockClass", blockClasses);
			of.addOption("tileEntityClass", tileClasses);
			of.addOption("autoDetectCuttableBlocks", autoDetect);
			
			blockClasses.values.addAll(Arrays.asList(props.getProperty("microblockTransformer.blockClasses").split(";")));
			tileClasses.values.addAll(Arrays.asList(props.getProperty("microblockTransformer.tileEntityClasses").split(";")));
			autoDetect.value = props.getProperty("autoDetectCuttableBlocks", "false").equals("true");
			
			of.write(configFile);
			
			oldConfigFile.delete();
		}
	}
	
	public static void load(File minecraftDir) {
		if(loaded)
			return;
		
		loaded = true;
		
		configDir = new File(minecraftDir, "config");
		if(!configDir.exists() && !configDir.mkdirs())
			throw new RuntimeException("Couldn't create directory: "+configDir);
		configFile = new File(configDir, "immibis-microblocks.cfg");
		
		try {
			updateConfig();
		} catch(IOException e) {
			throw new RuntimeException(e);
		}
		
		StringListOption blockClasses = new StringListOption(DEFAULT_BLOCK_CLASSES);
		StringListOption tileClasses = new StringListOption(DEFAULT_TILE_CLASSES);
		BooleanOption autoDetect = new BooleanOption(false);
		ItemAndMetaListOption cuttableBlocks = new ItemAndMetaListOption();
		BooleanOption enableNEIOption = new BooleanOption(true);
		
		of = new OptionsFile();
		of.addOption("blockClass", blockClasses);
		of.addOption("tileEntityClass", tileClasses);
		of.addOption("autoDetectCuttableBlocks", autoDetect);
		of.addOption("cuttableBlock", cuttableBlocks);
		of.addOption("enableNEI", enableNEIOption);
		
		if(configFile.exists()) {
			try {
				of.read(configFile);
			} catch(IOException e) {
				throw new RuntimeException(e);
			}
		}
		
		manualCuttableBlocks = cuttableBlocks.values;
		autoDetectCuttableBlocks = autoDetect.value;
		enableNEI = enableNEIOption.value;
		MicroblockSupporterTransformer.blockClasses.addAll(blockClasses.values);
		MicroblockSupporterTransformer.tileClasses.addAll(tileClasses.values);
		
		save();
	}
	
	public static void save() {
		of.<ItemAndMetaListOption>getOption("cuttableBlock").values = manualCuttableBlocks;
		of.<BooleanOption>getOption("autoDetectCuttableBlocks").value = autoDetectCuttableBlocks;
		try {
			of.write(configFile);
		} catch(IOException e) {
			throw new RuntimeException(e);
		}
	}
}
